@@ -8,58 +8,85 @@ import sys
 from attrdict import AttrDict
 import attrdict
 from getpass import getpass
-from multiprocessing import Process
-import multiprocessing as mp
-from websocket import _cookiejar
 
-
-mp.set_start_method('fork')
+__all__ = ['botlog', 'randomize_between', 'Bot']
 
 def botlog(message, error = False, _copy = False):
     ERROR = "\033[1;31;40m"
     NORMAL = "\033[m"
-    print((ERROR if error else NORMAL) + ("COPY: " if _copy else "") + message + NORMAL)
+    print(f'{ERROR if error else NORMAL}{"COPY: " if _copy else ""}{message}{NORMAL}')
     sys.stdout.flush()
 
+randomize_between = lambda *strings : random.choice(strings)
+
 class Bot:
-
-
-    def __init__(self, nick, room = "bots", owner = None, password = "", help = None, ping = None, important = False, killed = None, API_timeout = 10, advanced = False, function = None, sendable = False, _copy = False, _cookie = None, human = False):
-        parser = argparse.ArgumentParser(description=f'{nick}: A euphoria.io bot.')
-        parser.add_argument("--test", "--debug", "-t", help = "Used to debug dev builds. Sends bot to &test instead of its default room.", action = 'store_true')
-        parser.add_argument("--room", "-r", help = f"Set the room the bot will be placed in. Default: {room}", action = "store", default = room)
-        parser.add_argument("--password", "-p", help = "Set the password for the room the bot will be placed in.", action = "store", default = password)
-
-        args = parser.parse_args()
+    def __init__(self, nick, **kwargs):
+        vars = ["userEmail",
+            "userPassword",
+            "help",
+            "ping",
+            "killed",
+            "function",
+            "_cookie",
+            "human",
+            "_copy",
+            "sendable",
+            "APITimeout",
+            "advanced",
+            "owner",
+            "important"]
+        noneVars = ["userEmail",
+            "userPassword",
+            "help",
+            "ping",
+            "killed",
+            "function",
+            "_cookie"]
         self.nick = nick
-        self.room = args.room if args.test != True or _copy else "test"
         self.normname = re.sub(r"\s+", "", self.nick)
-        self.owner = owner if owner != None else "nobody"
-        self.normowner = re.sub(r"\s+", "", self.owner)
-        self.password = args.password
+        self.owner = "nobody"
         self.handlers = {}
         self.copies = []
-        self.help = help
-        self.ping = ping
-        self.important = important
-        self.killed = killed
-        self.API_timeout = API_timeout
-        self.advanced = advanced
-        self.function = function
-        self.sendable = sendable
-        self.copyProcesses = []
-        self.cookie = _cookie
+        self.important = False
+        self.APITimeout = 10
+        self.advanced = False
+        self.sendable = False
+        self.copyThreads = []
         self.conn = None
-        self.human = human
+        self.human = False
         self._regexes = {}
-        self._copy = _copy
-        if not _copy:
-            botlog("Debug: " + str(args.test), _copy = self._copy)
+        self._copy = False
+        for item in vars:
+            if item in kwargs:
+                self.setattr(item, kwargs[item])
+            elif item in noneVars:
+                self.setattr(item, None)
+        self.normowner = re.sub(r"\s+", "", self.owner)
+        self.parser = argparse.ArgumentParser(description=f'{nick}: A euphoria.io bot.' if 'description' not in kwargs else kwargs['description'])
+        self.args = self.parser.parse_args()
+        if 'customArgs' in kwargs and kwargs['customArgs']:
+            self.room = "bots" if "room" not in kwargs else kwargs["room"]
+            self.password = "" if "password" not in kwargs else kwargs["password"]
+        else:
+            self.add_command_line_arg(args = ["--test", "--debug", "-t"], help = "Used to debug dev builds. Sends bot to &test instead of its default room.", action = 'store_true')
+            self.add_command_line_arg(args = ["--room", "-r"], help = f"Set the room the bot will be placed in. Default: {'bots' if 'room' not in kwargs else kwargs['room']}", default = "bots" if "room" not in kwargs else kwargs["room"])
+            self.add_command_line_arg(args = ["--password", "-p"], help = "Set the password for the room the bot will be placed in.", default = "" if "password" not in kwargs else kwargs["password"])
+            self.room = self.args.room if self.args.test != True or _copy else "test"
+            self.password = self.args.password
+            if not self._copy:
+                botlog(f"Debug: {str(self.args.test)}", _copy = self._copy)
 
+    def add_command_line_arg(self, *args, **kwargs):
+        self.parser.add_argument(*args, **kwargs)
+        self.args = self.parser.parse_args()
+        if "name" in kwargs:
+            return self.args.kwargs["name"]
+        else:
+            return self.args
 
     def connect(self):
         self.conn = websocket.create_connection(f'wss://euphoria.io/room/{self.room}/ws{"?h=0" if not self.human else "?h=1"}', cookie = self.cookie, enable_multithread = True)
-        if self.cookie == None:
+        if self.cookie is None:
             self.cookie = self.conn.headers['set-cookie']
         hello = AttrDict(json.loads(self.conn.recv()))
         reply = AttrDict(json.loads(self.conn.recv()))
@@ -76,63 +103,37 @@ class Bot:
         return hello
 
     def send_msg(self, msgString, parent = None):
-        if re.search(r"^\[.+,.+\]$", msgString):
-            msgString = random.choice(msgString[1:-1].split(","))
         try:
             if isinstance(parent, AttrDict):
                 self.conn.send(json.dumps({'type': 'send', 'data': {'content': msgString, 'parent': parent.data.id}}))
-                reply = None
-                i = 0
-                for msgJSON in self.conn:
-                    i += 1
-                    msg = AttrDict(json.loads(msgJSON))
-                    if msg.type == "send-reply":
-                        reply = msg
-                        botlog(f'Message sent: {reply} replying to: {parent.data.id} by {parent.data.sender.name}', _copy = self._copy)
-                        return reply
-                    if i > self.API_timeout:
-                        botlog("send-reply API response not recorded. reason: too many events before send-reply.", True, _copy = self._copy)
-                        return None
-
             elif isinstance(parent, str):
                 self.conn.send(json.dumps({'type': 'send', 'data': {'content': msgString, 'parent': parent}}))
-                reply = None
-                i = 0
-                for msgJSON in self.conn:
-                    i += 1
-                    msg = AttrDict(json.loads(msgJSON))
-                    if msg.type == "send-reply":
-                        reply = msg
-                        botlog(f'Message sent: {reply} replying to: {parent}', _copy = self._copy)
-                        return reply
-                    if i > self.API_timeout:
-                        botlog("send-reply API response not recorded. reason: too many events before send-reply.", True, _copy = self._copy)
-                        return None
-            elif parent == None:
+            elif parent is None:
                 self.conn.send(json.dumps({'type': 'send', 'data': {'content': msgString}}))
-                reply = None
-                i = 0
-                for msgJSON in self.conn:
-                    i += 1
-                    msg = AttrDict(json.loads(msgJSON))
-                    if msg.type == "send-reply":
-                        reply = msg
-                        botlog(f'Message sent: {reply}', _copy = self._copy)
-                        return reply
-                    if i > self.API_timeout:
-                        botlog("send-reply API response not recorded. reason: too many events before send-reply.", True, _copy = self._copy)
-                        return None
-
             else:
                 raise BadMessageError(f'message not sent, because type of parent was {type(parent)}. \nParent printed: \n{parent}')
+            reply = None
+            i = 0
+            for msgJSON in self.conn:
+                i += 1
+                msg = AttrDict(json.loads(msgJSON))
+                if msg.type == "send-reply":
+                    reply = msg
+                    if isinstance(parent, AttrDict):
+                        botlog(f'Message sent: {reply} replying to: {parent.data.id} by {parent.data.sender.name}', _copy = self._copy)
+                    elif isinstance(parent, str):
+                        botlog(f'Message sent: {reply} replying to: {parent}', _copy = self._copy)
+                    elif parent is None:
+                        botlog(f'Message sent: {reply}', _copy = self._copy)
+                    return reply
+                if i > self.APITimeout:
+                    botlog("send-reply API response not recorded. reason: too many events before send-reply.", True, _copy = self._copy)
+                    return None
         except BadMessageError:
             pass
 
-
-
-
     def restart(self, msg = None, _copyStart = False):
-        if self.conn != None:
+        if self.conn is not None:
             self.conn.close()
             del self.conn
             botlog("restarting...", _copy = self._copy)
@@ -141,7 +142,6 @@ class Bot:
         self.connect()
         self.start()
 
-
     def start(self, function = None, advanced = False):
         if not (self.advanced or advanced):
             try:
@@ -160,24 +160,25 @@ class Bot:
             except KilledError:
                 pass
         else:
-            if self.function == None and function != None and callable(function):
+            if self.function is None and function is not None and callable(function):
                 self.function = function
             else:
-                botlog("No function passed to advanced bot's start function.", _copy = self._copy)
+                botlog("No function passed to advanced bot's start function. Using filler function", _copy = self._copy)
+                self.function = lambda msg: None
             try:
                 for msgJSON in self.conn:
                     msg = AttrDict(json.loads(msgJSON))
                     if msg.type == 'send-event' and msg.data.sender.name != self.nick:
-                        if re.search(f'^!kill @{self.normname}$', msg.data.content) != None:
+                        if re.search(f'^!kill @{re.escape(self.normname)}$', msg.data.content) is not None:
                             self._handle_kill(msg)
-                        if re.search(f'^!forcekill @{self.normname}$', msg.data.content) != None:
+                        if re.search(f'^!forcekill @{re.escape(self.normname)}$', msg.data.content) is not None:
                             self.kill()
-                        if re.search(f'^!restart @{self.normname}$', msg.data.content) != None:
+                        if re.search(f'^!restart @{re.escape(self.normname)}$', msg.data.content) is not None:
                             self.restart(msg = msg)
-                    if msg.type in self.handlers.keys():
+                    if msg.type in self.handlers:
                         self.handlers[msg.type](msg)
                     else:
-                        if function != None and callable(function):
+                        if function is not None and callable(function):
                             function(msg)
                         elif callable(self.function):
                             self.function(msg)
@@ -186,31 +187,29 @@ class Bot:
             except KilledError:
                 pass
 
-
-
     def _is_privileged(self, user):
-        if "is_manager" in user.keys() or user.name == self.owner:
+        if "is_manager" in user or user.name == self.owner:
             return True
         else:
             return False
 
     def set_regexes(self, regexes):
         self._regexes = regexes
-        if self.help == None:
-            if f"^!help @{self.normname}$" in regexes.keys():
-                self.help = regexes[f"^!help @{self.normname}$"]
+        if self.help is None:
+            if f"^!help @{re.escape(self.normname)}$" in regexes:
+                self.help = regexes[f"^!help @{re.escape(self.normname)}$"]
             else:
                 self.help = f"@{self.normname} is a bot made with Doctor Number Four's Python 3 bot library, DocLib (link: https://github.com/milovincent/DocLib) by @{self.normowner}.\nIt follows botrulez and does not have a custom !help message yet."
-        if self.ping == None:
-            if '^!ping$' in regexes.keys():
+        if self.ping is None:
+            if '^!ping$' in regexes:
                 self.ping = regexes['^!ping$']
-            elif f'^!ping @{self.normname}$' in regexes.keys():
-                self.ping = regexes[f'^!ping @{self.normname}$']
+            elif f'^!ping @{re.escape(self.normname)}$' in regexes:
+                self.ping = regexes[f'^!ping @{re.escape(self.normname)}$']
             else:
                 self.ping = "Pong!"
-        if self.killed == None:
-            if f"^!kill @{self.normname}$" in regexes.keys():
-                self.killed = regexes[f"^!kill @{self.normname}$"]
+        if self.killed is None:
+            if f"^!kill @{re.escape(self.normname)}$" in regexes:
+                self.killed = regexes[f"^!kill @{re.escape(self.normname)}$"]
             else:
                 self.killed = "/me has died."
 
@@ -218,20 +217,20 @@ class Bot:
         self.conn.send(json.dumps({'type': 'ping-reply', 'data': {'time': msg.data.time}}))
 
     def _handle_message(self, msg):
-        if re.search(f'^!kill @{self.normname}$', msg.data.content) != None:
+        if re.search(f'^!kill @{re.escape(self.normname)}$', msg.data.content) is not None:
             self._handle_kill(msg)
-        if re.search(f'^!forcekill @{self.normname}$', msg.data.content) != None:
+        if re.search(f'^!forcekill @{re.escape(self.normname)}$', msg.data.content) is not None:
             self.kill()
-        if re.search(f'^!restart @{self.normname}$', msg.data.content) != None:
+        if re.search(f'^!restart @{re.escape(self.normname)}$', msg.data.content) is not None:
             self.restart(msg = msg)
-        if re.search('^!ping$', msg.data.content) != None or re.search(f'^!ping @(?i){self.normname}$', msg.data.content) != None:
+        if re.search('^!ping$', msg.data.content) is not None or re.search(f'^!ping @(?i){re.escape(self.normname)}$', msg.data.content) is not None:
             self.send_msg(self.ping, parent = msg)
-        if re.search(f'^!help @(?i){self.normname}$', msg.data.content) != None:
+        if re.search(f'^!help @(?i){re.escape(self.normname)}$', msg.data.content) is not None:
             self.send_msg(self.help, parent = msg)
-        if re.search(f'^!send @(?i){self.normname} &(\w+)$', msg.data.content) != None and self.sendable:
-            self.move_to(re.search(f'^!send @(?i){self.normname} &(\w+)$', msg.data.content).group(0))
+        if re.search(f'^!send @(?i){re.escape(self.normname)} &(\w+)$', msg.data.content) is not None and self.sendable:
+            self.move_to(re.search(f'^!send @(?i){re.escape(self.normname)} &(\w+)$', msg.data.content).group(0))
         for regex, response in self._regexes.items():
-            if re.search(regex, msg.data.content) != None:
+            if re.search(regex, msg.data.content) is not None:
                 if callable(response):
                     result = response(self, msg, re.search(regex, msg.data.content))
                     if type(result) == str:
@@ -251,13 +250,12 @@ class Bot:
     def _handle_kill(self, msg):
         if self._is_privileged(msg.data.sender) or not self.important:
             self.send_msg(self.killed, msg)
-            for process in self.copyProcesses:
+            for process in self.copyThreads:
                 process.terminate()
             for copy in self.copies:
                 copy.send_msg("/me was killed via parricide")
                 copy.kill()
             self.kill()
-
         else:
             self.send_msg(f"Bot not killed because you are not a host or @{self.normowner}, and this bot is marked as important.\nFor emergencies, use !forcekill.", parent = msg)
 
@@ -276,15 +274,14 @@ class Bot:
                     botlog(f'Login unsuccessful. Reason: {reply.data.reason}', error = True, _copy = self._copy)
                     self._handle_auth(getpass("Enter the correct password: "))
                 break
-            if i > self.API_timeout:
+            if i > self.APITimeout:
                 botlog("Auth Error: auth-reply API response not recorded.", error = True, _copy = self._copy)
                 break
 
     def _handle_other(self, msg):
-        if msg.type in self.handlers.keys():
+        if msg.type in self.handlers:
             self.handlers[msg.type](msg)
 
-
     def set_nick(self, nick):
         self.conn.send(json.dumps({'type': 'nick', 'data': {'name': nick}}))
 
@@ -293,7 +290,7 @@ class Bot:
         raise KilledError(f'{self.normname} killed.')
 
     def get_userlist(self):
-        self.conn.send(json.dumps({'type': 'who', 'data': {}}))
+        self.conn.send(json.dumps({'type': 'who'}))
         reply = None
         i = 0
         for msgJSON in self.conn:
@@ -302,11 +299,10 @@ class Bot:
             if msg.type == "who-reply":
                 reply = msg
                 return reply.data.listing
-            if i > self.API_timeout:
+            if i > self.APITimeout:
                 botlog("who-reply API response not recorded. reason: too many events before who-reply.", error = True, _copy = self._copy)
                 return None
 
-
     def move_to(self, roomName, password = "", _copyStart = False):
         self.room = roomName
         self.password = password
@@ -321,36 +317,46 @@ class Bot:
             botlog(f"WARNING: handler for {eventString} not callable, handler not set.", error = True, _copy = self._copy)
 
     def set_handlers(self, handlers):
-        for eventString in handlers:
-            self.set_handler(eventString, handlers[eventString])
+        for eventString, function in handlers.items():
+            self.set_handler(eventString, function)
 
     def copy_to(self, roomName, password = None):
         copy = Bot(nick = self.nick,
-        room = roomName,
-        owner = self.owner,
-        password = self.password if password == None else password,
-        help = self.help,
-        ping = self.ping,
-        important = self.important,
-        killed = self.killed,
-        API_timeout = self.API_timeout,
-        advanced = self.advanced,
-        function = self.function,
-        sendable = self.sendable,
-        _copy = True,
-        _cookie = self.cookie,
-        human = self.human)
+            room = roomName,
+            owner = self.owner,
+            userEmail = self.userEmail,
+            userPassword = self.userPassword,
+            password = self.password if password is None else password,
+            help = self.help,
+            ping = self.ping,
+            important = self.important,
+            killed = self.killed,
+            APITimeout = self.APITimeout,
+            advanced = self.advanced,
+            function = self.function,
+            sendable = self.sendable,
+            _copy = True,
+            _cookie = self.cookie,
+            human = self.human)
         copy.set_handlers(self.handlers)
         copy.set_regexes(self._regexes)
-        p = Process(target = copy.move_to, kwargs = {"roomName" : roomName, "password" : password, "_copyStart" : True})
+        t = threading.Thread(target = copy.move_to)
         self.copies.append(copy)
-        self.copyProcesses.append(p)
-        p.start()
-        p.join(0)
+        self.copyThreads.append(t)
+        t.run(roomName = roomName, password = password, _copyStart = True)
+        t.join(0)
         return copy
 
-    def login(self, email = "", password = ""):
-        self.conn.send(json.dumps({'type': 'login', 'data': {'namespace' : "email", 'id' : email, 'password' : password}}))
+    def login(self, email = None, password = None, setDefaults = False):
+        if email is not None and password is not None:
+            self.conn.send(json.dumps({'type': 'login', 'data': {'namespace' : "email", 'id' : email, 'password' : password}}))
+            if setDefaults:
+                self.userEmail = email
+                self.userPassword = password
+        elif (self.userEmail is not None and email is None) or (self.userPassword is not None and password is None):
+            self.login(email = self.userEmail if email is None else email, password = self.userPassword if password is None else password, setDefaults = setDefaults)
+        else:
+            self.login(email = input("Enter your email: ") if email is None else email, password = getpass("Enter your password: ") if password is None else password, setDefaults = setDefaults)
         reply = None
         i = 0
         for msgJSON in self.conn:
@@ -363,16 +369,16 @@ class Bot:
                     self.account = reply.data.account_id
                 else:
                     botlog(f"Access denied. Reason: {reply.data.reason}", error = True, _copy = self._copy)
-                    self.login(email = input("Enter your email: "), password = getpass("Enter your password: "))
+                    self.login(email = input("Enter your email: "), password = getpass("Enter your password: "), setDefaults = setDefaults)
                 return reply
-            if i > self.API_timeout:
+            if i > self.APITimeout:
                 botlog("login-reply API response not recorded. reason: too many events before pm-initiate-reply.", error = True, _copy = self._copy)
                 return None
 
-    def initiate_pm(self, id, bot = None):
-        copy = Bot(nick = self.nick, room = "pmtest")
+    def initiate_pm(self, id, bot = None, room = "pmtest", hostEmail = None, hostPassword = None):
+        copy = Bot(nick = self.nick, room = room)
         copy.connect()
-        copy.login()
+        copy.login(email = hostEmail, password = hostPassword, setDefaults = False)
         copy.conn.close()
         del copy.conn
         copy.connect()
@@ -385,20 +391,19 @@ class Bot:
             if msg.type == "pm-initiate-reply":
                 copy.conn.close()
                 to = msg
-                if to.data != None:
-                    if bot != None:
-                        bot.copy_to("pm:" + to.data.pm_id)
+                if to.data is not None:
+                    if bot is not None:
+                        bot.copy_to(f"pm:{to.data.pm_id}")
                     else:
-                        self.copy_to("pm:" + to.data.pm_id)
+                        self.copy_to("pm:{to.data.pm_id}")
                 else:
                     botlog("Access Denied: This bot does not have permission to initiate PMs.", error = True, _copy = self._copy)
                     botlog(f"reply: {msg}", error = True, _copy = self._copy)
                 return to
-            if i > self.API_timeout:
+            if i > self.APITimeout:
                 botlog("pm-initiate-reply API response not recorded. reason: too many events before pm-initiate-reply.", error = True, _copy = self._copy)
                 return None
 
-
 class BotError(Exception):
     pass
 
